""" The graph function wraps an input class, adding properties accessing
methods decorated with arcbound.arc.

TODO: Add an option to show nodes without edges.
"""

import functools
from typing import Callable, Dict, List, Tuple, TypeVar

from .arc import Arc, arcs, auto_arcs
from .visualize import Digraph

ClassType = TypeVar("ClassType")
ArcboundGraph = TypeVar("ArcboundGraph", bound="ArcboundGraph")


def create_arcbound_graph(cls: ClassType) -> ArcboundGraph:
    """ Returns an ArcboundGraph object created using an input class.

    The ArcboundGraph object contains properties summarizing the attributes of
    the class decorated with arcbound.arc and a method for returning underlying
    functions generating the properties and methods.

    Args:
        cls: Class to graph methods and properties decorated with arcbound.arc.
    """
    class ArcboundGraph(object):
        """
        Properties:
            class_attrs: List of attribute names belonging to the input class.
            properties: Dictionary mapping methods decorated with both
                property and arcbound.arc to the property name.
            methods: Dictionary mapping methods decorated with arcbound.arc.
            nodes: Dictionary combining the properties and methods properties.
            arcs: Dictionary mapping dependencies of methods decorated with
                arcbound.arc.

        Methods:
            get_node: Returns as a curried function the decorated method with
                the instance variable already set.
        """
        @property
        @functools.lru_cache()
        def class_attrs(self) -> List[str]:
            """ Returns a list of attribute names belonging to the input class.
            """
            return dir(cls)

        @property
        @arcs(functions_list="class_attrs")
        def properties(self, functions_list: List[str]) -> Dict[str, Callable]:
            """ Returns a dictionary mapping property names to the method
            defining the property for methods decorated with arcbound.arc.
            """
            return {
                function_name: method.fget
                for function_name in functions_list
                for method in [getattr(cls, function_name)]
                if isinstance(method, property)
                if hasattr(method.fget, "arcs")
            }

        @property
        @arcs(functions_list="class_attrs")
        def methods(self, functions_list: List[str]) -> Dict[str, Callable]:
            """ Returns a dictionary mapping method names and methods for
            methods decorated with arcbound.arc.
            """
            return {
                function_name: method
                for function_name in functions_list
                for method in [getattr(cls, function_name)]
                if hasattr(method, "arcs")
            }

        @property
        @arcs(d1="properties", d2="methods")
        def nodes(
            self,
            d1: Dict[str, Callable],
            d2: Dict[str, Callable]
        ) -> Dict[str, Callable]:
            """ Combines the properties and methods into a single dictionary.
            """
            return {**d1, **d2}

        @property
        @auto_arcs()
        def edges(self, nodes: Dict[str, Callable]) -> Dict[str, Tuple[str]]:
            """ Returns a dictionary mapping dependencies by method/property.
            """
            return {
                name: {dep for dep in method.arcs}
                for name, method in nodes.items()
                if method.arcs
            }

        @auto_arcs()
        def get_node(self, k: str, nodes: Dict[str, Callable]) -> Callable:
            """ Returns the function defining the method or property.
            """
            def default_f(cls, *args, **kwargs) -> None:
                """ Dummy function returning None and indicating if the
                requested method is defined and/or decorated.
                """
                if k in dir(cls):
                    print("This method is not decorated.")
                else:
                    print("This method does not exist.")

                return None

            return nodes.get(k, default_f)

    return ArcboundGraph


def graph(cls: ClassType) -> Callable[[ClassType], ClassType]:
    """ Returns a class with properties and functions leveraging methods
    decorated with the arcbound.arcs function.

    Args:
        cls: Class to be decorated.

    Example:
        import arcbound as ab

        @ab.graph
        class test():
            def __init__(self, root_val: int) -> None:
                self.root = root_val
                return None

            @property
            @ab.arcs(x="root")
            def branch(self, x: int) -> int:
                return x * x

            @ab.arcs(x="branch", y="branch")
            def leaf(self, x: int, y: int) -> int:
                return x * y

            def twig(self, x: int, y: int) -> int:
                return x * y

        Test = test(5)

        Test.arcbound_graph.methods
        # {'leaf': <function __main__.test.leaf(self, x:int, y:int) -> int>}

        Test.arcbound_graph.properties
        # {'branch': <function __main__.test.branch(self, x:int) -> int>}

        Test.arcbound_graph.nodes
        # {'branch': <function __main__.test.branch(self, x:int) -> int>,
        #  'leaf': <function __main__.test.leaf(self, x:int, y:int) -> int>}

        Test.get_arcbound_node("branch")(10)
        # 100

        Test.get_arcbound_node("leaf")(2, 3)
        # 6

        Test.get_arcbound_node("twig")(2, 3)
        # This method is not decorated.

        Test.get_arcbound_node("nest")(10)
        # This method does not exist.
    """
    class wrapper_class(cls):
        """ Returns a class with properties and functions leveraging methods
        decorated with the arcbound.arcs function.

        Attributes:
            arcbound_graph: Returns an ArcboundGraph summarizing the input
                class's decorated methods.

        Functions:
            get_arcbound_node: Returns a curried function with the class's self
                provided as the class instance.
        """
        @property
        def arcbound_graph(self) -> ArcboundGraph:
            """ Returns an ArcboundGraph object used to summarize and interact
            with attributes decorated with arcbound.arcs.
            """
            return create_arcbound_graph(cls)()

        @arcs(arcbound_graph=Arc("arcbound_graph", silent=True))
        def get_arcbound_node(
            self,
            k: str,
            arcbound_graph: ArcboundGraph
        ) -> Callable:
            """ Returns the function defining the method or property with the
            instance variable already assigned.
            """
            f = arcbound_graph.get_node(k)

            return functools.partial(f, self)

        @arcs(graph=Arc("arcbound_graph", silent=True))
        def create_arcbound_graph(
            self,
            graph: ArcboundGraph,
            filename: str = "arcbound_graph.png",
            file_format: str = "png",
            **digraph_kwargs
        ) -> Digraph:
            """ Returns a graphviz-generated DiGraph made of the decorated
            methods.
            """
            return Digraph(
                deps_by_node=graph.edges,
                filename=filename,
                file_format=file_format,
                digraph_kwargs=digraph_kwargs
            ).graph

        def render_arcbound_graph(
            self,
            filename: str = "arcbound_graph",
            file_format: str = "png",
            directory: str = "./",
            **digraph_kwargs
        ) -> str:
            """ Save the graph. Returns the path to the saved file.
            """
            graph = self.create_arcbound_graph(
                filename=filename,
                file_format=file_format,
                **digraph_kwargs
            )

            return graph.render(filename=filename, directory=directory)

    return wrapper_class



""" The arc function wraps an input function to have default inputs set to
other attributes within the class and sets an attribute to the input function
documenting this dependency.
"""

import functools
import inspect
from typing import Any, Callable, Dict, Tuple, TypeVar, Union

import attr

ReturnType = TypeVar("ReturnType")
ClassType = TypeVar("ClassType")
FuncType = Callable[..., ReturnType]


@attr.s(auto_attribs=True, hash=False)
class Arc(object):
    """ Sets options on how an arc should be created.

    input_parameters:
        name: Name of the attribute to set as default.
        conditional: Function taking a class instance as an input, determining
            if the attribute should be set. The default function returns True.
        transform: Function that transforms the linked attribute before
            input to the decorated method.
        converter: Previous argument name for the transform input. Will be
            deprecated in the future.
        silent: Determines if the arc is documented.
        tag_only: Determines if the arc should be used only for documentation;
            will not set the default to the attribute if set to True.
    """
    name: str
    conditional: Callable[..., bool] = attr.Factory(lambda: lambda x: True)
    transform: Callable[..., Any] = attr.Factory(lambda: lambda x: x)
    converter: Callable[..., Any] = None
    silent: bool = False
    tag_only: bool = False

    def valid(self, cls: ClassType) -> bool:
        """ Determines if the default value should be assigned to the arc.
        """
        return self.conditional(cls)

    def value(self, cls: ClassType) -> Any:
        """ Return the value to be assigned as the default value in the
        function. If the converter argument is set, return the output of the
        function applied to the attribute.
        """
        initial_value = getattr(cls, self.name)

        if self.converter is None:
            value = self.transform(initial_value)

        else:
            print("Change converter to transform; will be deprecated.")
            value = self.converter(initial_value)

        return value


def arcs(
    auto_arcs: bool = False,
    silent_arcs: bool = False,
    arc_generator: Callable[..., Dict[str, Union[str, Arc]]] = None,
    arc_tags: Tuple[str] = (),
    extra_attribute_kwargs: dict = None,
    **attribute_kwargs
) -> Callable[[FuncType], FuncType]:
    """ Returns a function that sets instance attributes as default values.

    This decorator enables methods within a class to be constructed into a
    directed acyclic graph. Adds a 'deps' attribute to the returned function
    containing a dictionary defining the attributes to set as default values.

    Args:
        auto_arcs: Determines if arcs should automatically be created from all
            keywords in the input function.
        silent_arcs: Determines if the arcs are documented.
        arc_generator: Function defining arcs at runtime from the instance.
        arc_tags: Arcs to document without affecting the input function.
        extra_attribute_kwargs: Dictionary explicitly mapping keyword to
            attribute values to set as defaults in the input function to avoid
            potential namespace collisions.
        **attribute_kwargs: Keyword to attribute values to set as defaults in
            the input function.

    Example:
        import arcbound as ab

        class Test():
            def __init__(self, root_val: int) -> None:
                self.root = root_val
                return None

            @property
            @ab.arcs(x="root")
            def branch(self, x: int) -> int:
                return x * x

            @ab.arcs(x="branch", y="branch")
            def leaf(self, x: int, y: int) -> int:
                return x * y

        test = Test(5)
        test.branch # 25
        test.leaf() # 625
        test.leaf(x=10) # 250
        test.leaf(x=10, y=10) #100
    """
    def wrapper_factory(f: FuncType) -> FuncType:
        """ Returns a function with default values set to the object's
        attributes and adds the deps attribute to the wrapper function.

        Assumes that the first argument is the object-reference variable.

        If auto_arcs is true, attempts to set all arguments in the function
        with a default as the identically named attribute in the instance.

        Args:
            f: Function to decorate.
        """
        self_kw, *func_kws = inspect.getfullargspec(f).args

        # Combines the kwargs provided with an explicit dictionary argument
        # to avoid potential namespace collisions.
        initial_attribute_kwargs = dict(
            **attribute_kwargs,
            **(extra_attribute_kwargs or {})
        )

        assert all(
            isinstance(attribute, str) or isinstance(attribute, Arc)
            for attribute in initial_attribute_kwargs.values()
        ), "All arc definitions should either be a string or Arc object."

        @functools.wraps(f)
        def wrapper(*args, **kwargs) -> ReturnType:
            """ Returns the results of the input function with default values
            set by the attribute kwargs.
            """
            # Assign arguments without a keyword to the appropriate keywords.
            self, *non_self_args = args
            unused_kws = (k for k in func_kws if k not in kwargs)
            arg_kwargs = {k: v for k, v in zip(unused_kws, non_self_args)}

            # Generate attribute kwargs.
            generated_attribute_kwargs = (
                {} if arc_generator is None else
                arc_generator(self)
            )

            # Combines the kwargs provided with an explicit dictionary argument
            # to avoid potential namespace collisions.
            all_attribute_kwargs = dict(
                **initial_attribute_kwargs,
                **generated_attribute_kwargs,
            )

            # If auto_arcs is set to true, auto_kwargs maps the each argument
            # to the identically named attribute if it exists.
            if auto_arcs:
                auto_kwargs = {
                    k: getattr(self, k)
                    for k in func_kws
                    if k not in all_attribute_kwargs
                    if hasattr(self, k)
                }

            else:
                auto_kwargs = {}

            unpacked_attribute_kwargs = {
                k: (
                    getattr(self, attribute) if isinstance(attribute, str) else
                    attribute.value(self)
                )
                for k, attribute in all_attribute_kwargs.items()
                if (
                    isinstance(attribute, str)
                    or (attribute.valid(self) & (not attribute.tag_only))
                )
            }

            combined_kwargs = {
                **auto_kwargs,
                **unpacked_attribute_kwargs,
                **kwargs,
                **arg_kwargs
            }

            return f(self, **combined_kwargs)

        if silent_arcs:
            tagged_arcs = ()

        else:
            # Set edges even if conditionals were not met.
            if auto_arcs:
                auto_arc_tags = tuple(
                    k
                    for k in func_kws
                    if k not in initial_attribute_kwargs
                )

            else:
                auto_arc_tags = ()

            unpacked_arc_tags = tuple(
                attribute if isinstance(attribute, str) else
                attribute.name
                for k, attribute in initial_attribute_kwargs.items()
                if isinstance(attribute, str) or (not attribute.silent)
            )

            tagged_arcs = tuple(
                auto_arc_tags + unpacked_arc_tags + arc_tags
            )

        wrapper.arcs = tagged_arcs

        return wrapper

    return wrapper_factory


def arc(**attribute_kwargs) -> Callable[[FuncType], FuncType]:
    """ Placeholder for arcs. Will be deprecated on release.
    """
    return arcs(**attribute_kwargs)


def auto_arcs() -> Callable[[FuncType], FuncType]:
    """ Syntactical sugar for arcs(auto_arcs=True).
    """
    return arcs(auto_arcs=True)



""" The requires decorator checks if a package is available and raises an error
or sends a warning otherwise.
"""

import functools
import importlib.util
from typing import Callable, TypeVar, Union
import warnings

ClassType = TypeVar("ClassType")


def requires(
    package: str,
    message: str = None,
    warn_only: bool = False,
) -> Callable[[ClassType], ClassType]:
    """ Returns the input object with an added check that the input package
    is installed at instantiation.

    Args:
        package: Name of package to check for.
        message: Message to print out if package is not installed.
        warn_only: Flag determining if import errors should be downgraded to
            a warning.
    """
    def wrapper_factory(
        obj: Union[ClassType, Callable]
    ) -> Union[ClassType, Callable]:
        """ Returns the input class with an added check that a specified package
        is installed at instantiation.

        Args:
            obj: Class or function to be modified with a package check.
        """
        if callable(obj):
            @functools.wraps(obj)
            def obj_with_requires(*args, **kwargs):
                """ Adds a package check before executing the function.
                """
                library_loaded = importlib.util.find_spec(package)

                if (not library_loaded) & (not warn_only):
                    default_error = f"{obj.__name__} requires {package}."
                    raise ImportError(message or default_error)

                elif (not library_loaded):
                    default_warning = f"{package} not installed, proceeding."

                    warnings.warn(message or default_warning, ImportWarning)

                return obj(*args, **kwargs)

        else:
            class obj_with_requires(obj):
                """ Adds a package check to the __init__ method of the parent
                class.
                """
                __doc__ = obj.__doc__

                def __init__(self, *args, **kwargs):
                    """ First checks if the specified package is installed then
                    proceeds with the parent's initialization.

                    Raises:
                        ImportError: The required package is not found.
                    """
                    library_loaded = importlib.util.find_spec(package)

                    if (not library_loaded) & (not warn_only):
                        default_error = f"{obj.__name__} requires {package}."
                        raise ImportError(message or default_error)

                    elif (not library_loaded):
                        default_warning = f"""
                            {package} not installed, proceeding.
                            """

                        warnings.warn(
                            message or default_warning,
                            ImportWarning
                        )

                    super().__init__(*args, **kwargs)

        return obj_with_requires

    return wrapper_factory



"""
Initializes the arcbound package. Load classes and functions into the arcbound
namespace.
"""

from .arc import arc, arcs, Arc, auto_arcs  # noqa: F401
from .graph import graph  # noqa: F401
from .requires import requires  # noqa: F401



""" Visualizes arcbound graphs.

Requires graphviz.
"""

from typing import Dict, Set

import attr

try:
    import graphviz
except ImportError:
    pass

from .arc import arcs, auto_arcs
from .requires import requires


@requires("graphviz")
@attr.s(auto_attribs=True)
class Digraph(object):
    """ Creates a digraph from an input dictionary mapping dependencies to
    nodes.
    """
    deps_by_node: Dict[str, Set[str]]
    filename: str = "arcbound_digraph"
    file_format: str = "png"
    digraph_kwargs: dict = attr.Factory(dict)

    ###########################################################################
    # Set up the nodes and edges.
    ###########################################################################

    @property
    @auto_arcs()
    def nodes(self, deps_by_node: Dict[str, Set[str]]) -> Set[str]:
        """ Returns the nodes in the graph.
        """
        return (
            set(deps_by_node)
            | {node for nodes in deps_by_node.values() for node in nodes}
        )

    @property
    @auto_arcs()
    def edges(self, deps_by_node: Dict[str, Set[str]]) -> Set[str]:
        """ Returns the edges in the graph.
        """
        return {
            f"{node}{dep}"
            for node, deps in deps_by_node.items()
            for dep in deps
        }

    ###########################################################################
    # Draw the graph.
    ###########################################################################

    @property
    @auto_arcs()
    def blank_graph(
        self,
        graph_name: str = "test",
        filename: str = "test",
        file_format: str = "png",
        digraph_kwargs: dict = None
    ) -> graphviz.Digraph:
        """ Returns an instantiated graphviz Digraph object.

        Args:
            graph_name: Graph name used in the source code.
            filename: Filename to save the output to.
            file_format: How to render the output format.
            digraph_kwargs (dict): Additional kwargs to pass to the Digraph
                instantiation.
        """
        return graphviz.Digraph(
            name=graph_name,
            filename=filename,
            format=file_format,
            **digraph_kwargs
        )

    @property
    @arcs(dag="blank_graph", deps_by_node="deps_by_node")
    def graph(
        self,
        dag: graphviz.Digraph,
        deps_by_node: Dict[str, Set[str]]
    ) -> graphviz.Digraph:
        """ Returns a graphviz Digraph object with the nodes and edges defined
        in the arcbound graph.
        """
        for node in deps_by_node:
            dag.node(name=node, label=node)

        for node, deps in deps_by_node.items():
            for dep in deps:
                dag.edge(dep, node)

        return dag



""" Arcbound is a collection of tools used to develop and maintain data-driven
Python code, organizing attributes and methods of a class into a graph.
"""

import os
import re

from setuptools import setup, find_packages

__author__ = "Jason Hwang"
__email__ = "@".join(("jhwang.astro", "gmail.com"))
__url__ = "https://github.com/JHwangAstro/arcbound"
__description__ = "Collection of tools to arrange bound methods into a graph."


def get_version(*paths) -> str:
    """ Returns the version number by looking in the file paths specified.
    Assumes that the paths provided are relative.

    Args:
        paths: Relative path to the file containing the version.

    Raises:
        RuntimeError: If the version string is unable to be found.
    """
    setup_file_path = os.path.abspath(os.path.dirname(__file__))
    init_path = os.path.join(setup_file_path, *paths)

    with open(init_path) as init_file:
        init_blob = init_file.read()

    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]",
        init_blob,
        re.M
    )

    if version_match is None:
        raise RuntimeError("Unable to find version string.")

    else:
        version = version_match.group(1)

    return version


this_directory = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

install_requires = [
    "attrs>=19.3.0"
]

extras_require = {
    # Required to visualize methods graph.
    "graphviz": ["graphviz==0.13.2"]
}

setup(
    name="arcbound",
    version=get_version("arcbound", "release.py"),
    description=__description__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=__author__,
    author_email=__email__,
    url=__url__,
    python_requires=">=3.6",
    install_requires=install_requires,
    extras_require=extras_require,
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent"
    ]
)
